from bertdotbill.logger import Logger
from bertdotbill.defaults import gui_dirname
import os
import sys

logger = Logger().init_logger(__name__)
is_frozen = getattr(sys, 'frozen', False)
if is_frozen:
  my_file_name = os.path.basename(sys.executable)
  project_root = os.path.dirname(os.path.abspath(sys.executable))  
else:
  my_file_name = __file__
  project_root = os.path.dirname(my_file_name)

def get_entrypoint():
  index_file_path_found = False
  logger.info('Determining path to index.html')
  if is_frozen: # Check for frozen pyinstaller app
    logger.debug('Detected installation type is "frozen"')
    html_index_file_path_relative = './%s/index.html' % gui_dirname
    html_index_file_path = os.path.join(project_root, html_index_file_path_relative)
    logger.info('Checking %s' % html_index_file_path)
    if os.path.exists(html_index_file_path):
      index_file_path_found = True
      logger.info('Found %s' % html_index_file_path)
      return html_index_file_path
    else: # Check for frozen py2app
      html_index_file_path_relative = '../Resources/%s/index.html' % gui_dirname
      html_index_file_path = os.path.join(project_root, html_index_file_path_relative)
      logger.info('Checking %s' % html_index_file_path)
      if os.path.exists(html_index_file_path):
        index_file_path_found = True
        logger.info('Found %s' % html_index_file_path)
        return html_index_file_path_relative
      else:
        logger.error('%s not found' % html_index_file_path)
  else: # Check for unfrozen development app
    import sysconfig
    root_package_name = __name__.split('.')[0]
    site_packages_path = sysconfig.get_paths()['purelib']
    site_scripts_path = sysconfig.get_paths()['scripts']
    root_package_path = os.path.join(site_packages_path, root_package_name)
    if os.path.isdir(root_package_path):
      logger.debug('Detected installation type is "pip"')
      html_index_file_path_relative = './%s/index.html' % gui_dirname
      html_index_file_path = os.path.join(site_scripts_path, gui_dirname, 'index.html')
    else:
      logger.debug('Detected installation type is "development"')
      html_index_file_path_relative = '../%s/index.html' % gui_dirname
      html_index_file_path = os.path.join(project_root, html_index_file_path_relative)
    logger.info('Checking %s' % html_index_file_path)
    if os.path.exists(html_index_file_path):
      index_file_path_found = True
      logger.info('Found %s' % html_index_file_path)
      return html_index_file_path_relative
    else:
      logger.error('%s not found' % html_index_file_path)
  if not index_file_path_found:
    raise Exception('No index.html found')


import base64
from bertdotbill.lessons import Lessons
from bertdotbill.logger import Logger
from bertdotbill.terminals import Terminals
import sys

logger = Logger().init_logger(__name__)

class UI():

  def __init__(self, **kwargs):
    self.settings = kwargs['settings']
    self.args = kwargs['args']
    # Initialize Lesson Loader
    self.lessons = Lessons(
    settings=self.settings, 
    args=self.args)
    # Initialize Terminals
    self.terminals = Terminals(
    settings=self.settings, 
    args=self.args)    


  def init_terminal(self, **kwargs):
    
    webview = kwargs['webview']
    section = kwargs['section']
    set_function = kwargs['set_function']
    url = self.terminals.load(section)
    webview.windows[0].evaluate_js(
      'window.pywebview.state.%s("%s")' % (set_function, url)
    )

  def initialize(self, webview):
    if len(webview.windows) > 0:
      logger.info('Initializing UI View')

      self.init_terminal(
        webview = webview,
        section = 'rightpane',
        set_function = 'setRightPaneURL'
      )

      self.init_terminal(
        webview = webview,
        section = 'footer',
        set_function = 'setfooterURL'
      )      

      if not self.args.no_init:
        webview.windows[0].evaluate_js('window.pywebview.state.loadLesson("")')
        if not self.args.no_init_lesson:
          webview.windows[0].evaluate_js('window.pywebview.state.setLessons("%s")' % self.lessons.load())
    else:
      logger.info('Skipped Lesson View Initialization')


from bertdotbill.defaults import default_terminal_address
from bertdotbill.defaults import terminal_check_timeout
from bertdotbill.logger import Logger
import socket
from bertdotconfig import Struct
from bertdotbill.config import SuperDuperConfig

logger = Logger().init_logger(__name__)

class Terminals():

  def __init__(self, **kwargs):
    self.settings = kwargs['settings']
    self.args = kwargs['args']
    self.superconf = SuperDuperConfig(verify_tls=self.args.verify_tls)
    self.default_console_address = self.superconf.get(
      self.settings,'terminals.default.address', 
      default_terminal_address
    )
    self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

  def load(self, section):

    terminals = self.settings.get('terminals', {})
    
    if terminals:
      terminal_address = terminals.get(
        section,{}).get('address', self.default_console_address)
    else:
      terminal_address = self.default_console_address
    logger.info('Terminal address for section {s}: {a}'.format(
      s=section, a=terminal_address
    ))
    
    return terminal_address


from bertdotbill.auth import get_credential
from bertdotbill.defaults import settings as default_settings
from bertdotbill.logger import Logger
import os
from bertdotconfig import SuperDuperConfig

logger = Logger().init_logger(__name__)

def read_config(**kwargs):

  args = kwargs['args']
  verify_tls = kwargs.get('verify_tls', False)

  try:
    script_dir = os.path.dirname(os.path.abspath(my.invocation.path))
  except NameError:
    script_dir = os.path.dirname(os.path.abspath(__file__))
  if not args.config_file:
    logger.debug('No config file specified')
  default_config_file_name = 'bill.config.yaml'
  default_config_file_search_paths = [
  os.path.abspath(os.path.join(script_dir)),
  "~/.bill"
  ]
  config_file_path = args.config_file
  # Initialize Config Module
  superconf = SuperDuperConfig(
    extra_config_search_paths=default_config_file_search_paths,
    verify_tls=verify_tls
  )
  # Initialize App Config
  initial_data = {
  'environment': os.environ,
  'get_credential': get_credential
  }  
  settings = superconf.load_config(
    config_file_name=default_config_file_name,
    templatized=args.config_file_templatized,
    initial_data=initial_data
  )
  if not settings:
    settings = default_settings  
  external_configs = settings.get('external_configs', [])
  for external_config in external_configs:
    if isinstance(external_config, dict):
      config_uri = external_config.get('uri')
      if config_uri:
        config_uri_username = external_config.get('auth',{}).get('username')
        config_uri_password = external_config.get('auth',{}).get('password')
        external_settings = superconf.load_config(
          config_file_uri=config_uri,
          auth_username=config_uri_username,
          auth_password=config_uri_password,
          templatized=args.config_file_templatized,
          initial_data=initial_data
        )
        if external_settings:
          settings = superconf.Merge(settings, external_settings)
  return superconf, settings


from bertdotbill.defaults import app_name
from bertdotbill.api import API
from bertdotbill.entrypoint import get_entrypoint
from bertdotbill.ui import UI
from bertdotbill.cli import parse_args
from bertdotbill.config import read_config
from bertdotbill.logger import Logger
import sys
import webview

# Read command-line args
args = parse_args()
# Initialize logging facility
logger = Logger().init_logger(__name__)
# Initialize Config Reader
superconf, settings = read_config(
  args=args, verify_tls=args.verify_tls
  )
# Initialize UI
ui = UI(settings=settings, args=args)
# Get html entrypoint
entry = get_entrypoint()

def main():

  # Load UI
  if not args.no_ui:
    logger.info('UI entrypoint is %s' % entry)
    window = webview.create_window(
      app_name,
      entry,
      js_api=API(webview, superconf, args, settings),
      width=args.window_width,
      height=args.window_height,      
      text_select=True)
    # Start the webview object, with initial function lessons.initialize
    # being called, with args webview,
    webview.start(ui.initialize, args=[webview], gui=args.webengine, debug=args.debug)
  else:
    # Run via console only
    lesson_url = args.lesson_url
    if not lesson_url:
      logger.error('When operating in noui mode, ensure you specify a lesson url')
      sys.exit(1)
    js_api=API(webview, superconf, args, settings)
    js_api.load_lesson(lesson_url, no_ui=True, norender_markdown=args.norender_markdown)

if __name__ == '__main__':
  main()


import base64
import os
import markdown
from jinja2 import Template
from bertdotbill.commands import launch
from bertdotbill.logger import Logger
from bertdotwebadapter import WebAdapter

logger = Logger().init_logger(__name__)

class API:

  def __init__(self, webview, superconf, args, settings, **kwargs):
    self.settings = settings
    self.webview = webview
    self.webadapter = WebAdapter(fail_on_errors=True, verify_tls=args.verify_tls)
    self.global_username = os.environ.get('GLOBAL_USERNAME') or args.username or superconf.get(self.settings,'auth.global.username')
    self.global_password = os.environ.get('GLOBAL_PASSWORD') or args.password or superconf.get(self.settings,'auth.global.password')
    
  def log(self, level, message):
    if level == 'error':
      logger.error(message)
    elif level == 'debug':
      logger.debug(message)
    else:
      logger.info(message)

  def fullscreen(self):
    self.webview.windows[0].toggle_fullscreen()

  def launch_command(self, command_key, args=None):
    launch(self.settings, command_key, args)

  # TODO: Move all lessons-related functions
  # to the lessons package file
  def encode_lesson(self, rendered_lesson):
      rendered_lesson_bytes = rendered_lesson.encode("ascii")
      encoded_lesson = base64.b64encode(rendered_lesson_bytes)      
      return encoded_lesson.decode("utf-8")

  def render_lesson(self, git_url, lesson_content, norender_markdown=False):

    initial_data = {
    'environment': os.environ
    }
    lesson_content = str(lesson_content)
    lesson_content = lesson_content.strip()
    lesson_template = Template(lesson_content)
    try:
      rendered_lesson = lesson_template.render(
        session=initial_data
      )
    except Exception as e:
      err = str(e)
      logger.error('I had trouble rendering the lesson at %s, error was %s' % (git_url, err))
      rendered_lesson = ('''
        <div class="lesson-error-container">
          <div class="lesson-error-text">
          Error in rendering lesson at %s:<br /> %s
          </div>
        </div>
        ''' % (git_url, err)
      )
      return rendered_lesson
    if norender_markdown:
      return rendered_lesson
    else:
      rendered_lesson = markdown.markdown(rendered_lesson).replace('<code>','<code class="clickable-code">')
      return rendered_lesson

  # TODO: git_url should be renamed to lesson_url
  def load_lesson(self, git_url, no_ui=False, norender_markdown=False):
    git_url = os.environ.get('GIT_URL') or git_url
    res_ok = False
    # TODO: Employ per-lesson credentials
    if not no_ui and len(self.webview.windows) > 0:
      try:
        res = self.webadapter.get(git_url, 
          username=self.global_username,
          password=self.global_password,        
          cache_path='.')
        res_ok = True
      except Exception as e:
        err = str(e)
        logger.error('I had trouble retrieving the lesson at %s, error was %s' % (git_url, err))
        html_err_message = '''
          <div class="lesson-error-container">
            <div class="lesson-error-text">
            I had trouble retrieving the lesson at %s<br />
            Error was: %s<br />
            </div>
          </div>
        ''' % (git_url, err)
        encoded_lesson = self.encode_lesson(html_err_message)
      if res_ok:
        lesson_content = res
        logger.info('Attempting to render and encode lesson at %s' % git_url)
        rendered_lesson = self.render_lesson(git_url, lesson_content, norender_markdown=norender_markdown)
        logger.debug(rendered_lesson)
        try:
          encoded_lesson = self.encode_lesson(rendered_lesson)
        except Exception as e:
          err = str(e)
          logger.error('I had trouble encoding the lesson at %s' % git_url, err)
          html_err_message = '''
            <div class="lesson-error-container">
              <div class="lesson-error-text">
              I had trouble encoding the lesson at %s<br />
              Error was: %s
              </div>
            </div>
          ''' % (git_url, err)
          encoded_lesson = self.encode_lesson(html_err_message)
      try:
        # Load the lesson content
        self.webview.windows[0].evaluate_js('window.pywebview.state.loadLesson("%s")' % encoded_lesson)
        # Add click events to all elements with class 'clickable-code'
        self.webview.windows[0].evaluate_js('window.pywebview.state.addClickEvents()')
      except Exception as e:
        self.webview.windows[0].evaluate_js('alert("Encountered an error: %s")' % e) 
    elif no_ui:
      res = self.webadapter.get(git_url, 
        username=self.global_username,
        password=self.global_password,        
        cache_path='.')
      lesson_content = str(res)
      rendered_lesson = self.render_lesson(lesson_content, norender_markdown=norender_markdown)
      print(rendered_lesson)

  def save_content(self, content):
    filename = self.webview.windows[0].create_file_dialog(self.webview.SAVE_DIALOG)
    if not filename:
        return

    with open(filename, 'w') as f:
        f.write(content)

  def ls(self):
    return os.listdir('.')


import argparse
from bertdotbill.defaults import app_name

from bertdotbill.defaults import webengine_choices

def parse_args(**kwargs):

  parser = argparse.ArgumentParser(description=app_name)
  parser.add_argument('--username', '-u', help="Username, if the URL requires authentication")
  parser.add_argument('--password', '-p', help="Password, if the URL requires authentication")
  parser.add_argument('--lesson-url', '-url', help="The URL for the lesson definition")
  parser.add_argument('--webengine', '-g', choices=webengine_choices, help="The Web Renderer")
  parser.add_argument('--window-width', '-W', default=1024, type=int, help="App Window Width")
  parser.add_argument('--window-height', '-H', default=600, type=int, help="App Window Height")
  parser.add_argument('--config-file', '-f', help="Path to app configuration file")
  parser.add_argument('--config-file-templatized', '-fT', action='store_true', default=True, help="Render configuration via jinja2 templating")
  parser.add_argument('--no-ui', '-noui', action='store_true', help="Don't launch the UI")
  parser.add_argument('--no-init', '-noinit', action='store_true', help="Skip all UI init steps")
  parser.add_argument('--no-init-lesson', '-noinitl', action='store_true', help="Skip UI init Lesson steps")
  parser.add_argument('--debug', action='store_true')
  parser.add_argument('--verify-tls', action='store_true', help='Verify SSL cert when downloading web content', default=False)
  parser.add_argument('--norender-markdown', '-nomarkdown', action='store_true')
  return parser.parse_args()


import base64
from bertdotbill.defaults import default_terminal_address
from bertdotbill.logger import Logger
from bertdotbill.config import SuperDuperConfig
import sys

logger = Logger().init_logger(__name__)

class Lessons():

  def __init__(self, **kwargs):
    self.settings = kwargs['settings']
    self.args = kwargs['args']
    self.superconf = SuperDuperConfig(verify_tls=self.args.verify_tls)

  def initialize(self, webview):
    if len(webview.windows) > 0:
      logger.info('Initializing Lesson View')
      default_console_address = self.superconf.get(
        self.settings,'terminals.default.address', default_terminal_address
      )
      terminals = self.settings.get('terminals', {})
      if terminals:
        terminal_address = terminals.get(
          'footer',{}).get('address', default_console_address)
      else:
        terminal_address = default_console_address
      webview.windows[0].evaluate_js('window.pywebview.state.setiFrameURL("%s")' % terminal_address)
      if not self.args.no_init:
        webview.windows[0].evaluate_js('window.pywebview.state.loadLesson("")')
        if not self.args.no_init_lesson:
          webview.windows[0].evaluate_js('window.pywebview.state.setLessons("%s")' % self.load())
    else:
      logger.info('Skipped Lesson View Initialization')

  def load(self):

    logger.info('Loading available topics')
    topics = self.settings.get('topics', {})
    lessons_html = ''
    available_lesson_count = 0
    if isinstance(topics, dict):
      lessons_html = '<a href="#">Available Lessons</a>\n'
      lessons_html += "<ul class='nav first'>"
      for lesson_name, lesson_data in topics.items():
        lessons_html += '<li>%s\n' % lesson_name
        if isinstance(lesson_data, dict):
          lessons = lesson_data.get('lessons', [])
          lessons_html += '<ul class="nav">\n'
          if isinstance(lessons, list):
            for lesson in lessons:
              if isinstance(lesson, dict):
                topic_name = lesson.get('name','')
                topic_url = lesson.get('url','#')
                lessons_html += '<li><button onclick="window.pywebview.api.load_lesson(\'%s\')">%s</button></li>\n' % (topic_url, topic_name)
            lessons_html += '</ul>\n</li>'
    if sys.version_info[0] >= 3:
        b64_lessons = base64.standard_b64encode((lessons_html).encode()).decode()
    else:
        b64_lessons = base64.standard_b64encode(lessons_html)
    logger.info('Successfully loaded available lessons')
    return b64_lessons


from bertdotbill.logger import Logger
import os
import sys

logger = Logger().init_logger(__name__)

def run_command(command, args=None):
  logger.debug('Running command %s' % command)
  os.system(command)

def launch(settings, command_key, args=None):
    logger.info("Launching command mapped to '%s'" % command_key)
    commands = settings.get('commands',{})
    if commands.get(sys.platform,{}).get(command_key):
      command = commands.get(sys.platform,{}).get(command_key,{}).get('cmd')
    else:
      command = commands.get(command_key,{}).get('cmd')
    if command:
      run_command(command, args)
    else:
      logger.warning('No command key found for %s' % command_key)


# Set App Name
app_name = "Bert's Interactive Lesson Loader (BILL)"
gui_dirname = 'bill.gui'
default_terminal_address = 'ws://127.0.0.1:5000/'
terminal_check_timeout = 5
webengine_choices = [
  'gtk',
  'qt',
  'edgechromium',
  'edgehtml',
  'mshtml',
  'cef'
]
settings = {}


import keyring
from bertdotconfig import Struct

def get_credential(entry):
  credential_data = keyring.get_credential(entry, None)
  credential = {
    'username': credential_data._username,
    'password': credential_data._password,
  }
  return Struct(credential)